from uuid import uuid4

import onnx
from .proto import get_maximum_opset_supported
from ._parse import parse_aad
from . import operator_converters
from . import shape_calculators
from onnxconverter_common.topology import convert_topology

def convert(model, name=None, initial_types=None, doc_string='', target_opset=None,
	targeted_onnx=onnx.__version__, custom_conversion_functions=None,
	custom_shape_calculators=None):

	if initial_types is None:
		raise ValueError('Initial types are required. See usage of '
				'convert(...) in aad2onnx.convert for details')

	if name is None:
		name = str(uuid4().hex)

	target_opset = target_opset if target_opset else get_maximum_opset_supported()

	topology = parse_aad(model, initial_types, target_opset, custom_conversion_functions, custom_shape_calculators)
	topology.compile()

	onnx_model = convert_topology(topology, name, doc_string, target_opset, targeted_onnx)
	return onnx_model

def to_onnx(model, X=None, name=None, initial_types=None, target_opset=None,
	targeted_onnx=onnx.__version__, custom_conversion_functions=None,
	custom_shape_calculators=None):

	from skl2onnx.algebra.type_helper import guess_initial_types

	if name is None:
		name = "ONNX(%s)" % model.__class__.__name__
	initial_types = guess_initial_types(X, initial_types)

	return convert(model, name=name, initial_types=initial_types,
		target_opset=target_opset, targeted_onnx=targeted_onnx,
		custom_conversion_functions=custom_conversion_functions,
		custom_shape_calculators=custom_shape_calculators)



from ad_examples.aad.aad_globals import get_aad_option_list, AAD_IFOREST, AadOpts
from ad_examples.aad.aad_support import get_aad_model
from aad2onnx import convert, to_onnx
import onnx
import onnxruntime as rt
import numpy as np
import unittest

def createAADOpts():
	rng = np.random.RandomState(42)

	parser = get_aad_option_list()
	args = parser.parse_args([])
	args.keep = None
	args.sparsity = np.nan
	args.detector_type = AAD_IFOREST

	opts = AadOpts(args)

	return opts, rng

class ConvertTest(unittest.TestCase):
	@classmethod
	def setUpClass(cls):
		cls._opts = createAADOpts()

	def createAADModel(self, x):
		model = get_aad_model(x, *self._opts)
		model.fit(x)
		model.init_weights()

		return model

	def evaluate_onnx(self, o, data):
		sess = rt.InferenceSession(o.SerializeToString())
		input_name = sess.get_inputs()[0].name
		label_name = 'score'
		scores_onnx = sess.run([label_name], {input_name: data.astype(np.float32)})[0].reshape(-1)

		return scores_onnx

	def test_convert1(self):
		x = np.random.normal(size=(20, 4)).astype(np.float32)

		model = self.createAADModel(x)
		scores_model = model.get_score(model.transform_to_ensemble_features(x)).astype(np.float32)

		o = to_onnx(model, x)
		onnx.checker.check_model(o)
		scores_onnx = self.evaluate_onnx(o, x)

		np.testing.assert_allclose(scores_onnx, scores_model, 1e-5)



from onnx import defs
from onnxconverter_common import onnx_ex

def get_maximum_opset_supported():
	from . import __max_supported_opset__
	return min(__max_supported_opset__, onnx_ex.get_maximum_opset_supported())



from onnxconverter_common.registration import register_converter


def get_default_attribute_pairs():
	attrs = {}
	attrs['post_transform'] = 'NONE'
	attrs['n_targets'] = 1
	attrs['nodes_treeids'] = []
	attrs['nodes_nodeids'] = []
	attrs['nodes_featureids'] = []
	attrs['nodes_modes'] = []
	attrs['nodes_values'] = []
	attrs['nodes_truenodeids'] = []
	attrs['nodes_falsenodeids'] = []
	attrs['target_treeids'] = []
	attrs['target_nodeids'] = []
	attrs['target_ids'] = []
	attrs['target_weights'] = []
	return attrs

def get_decision_path(node_id, parents, node_regions):
	decision_path = [node_regions[node_id],]

	next_id = parents.get(node_id)
	# next_id = 0 is the root which is not a region for AAD
	while next_id > 0:
		decision_path.append(node_regions[next_id])
		next_id = parents.get(next_id)

	return decision_path

def get_leaf_weight_only(tree_id, node_id, model):
	region_id = model.all_node_regions[tree_id][node_id]
	score = model.get_region_score_for_instance_transform(region_id, 1)

	return model.w[region_id] * score

def get_leaf_weight_full(tree_id, node_id, parents, model):
	node_regions = model.all_node_regions[tree_id]
	decision_path = get_decision_path(node_id, parents, node_regions)
	k = len(decision_path)

	return sum([model.w[region_id] * model.get_region_score_for_instance_transform(region_id, k) for region_id in decision_path])

def get_leaf_weight(tree_id, node_id, parents, model):
	if model.add_leaf_nodes_only:
		return get_leaf_weight_only(tree_id, node_id, model)
	else:
		return get_leaf_weight_full(tree_id, node_id, parents, model)

def add_node(attr_pairs, tree_id, node_id, feature_id, mode, value, true_child_id, false_child_id):
	attr_pairs['nodes_treeids'].append(tree_id)
	attr_pairs['nodes_nodeids'].append(node_id)
	attr_pairs['nodes_featureids'].append(feature_id)
	attr_pairs['nodes_modes'].append(mode)
	attr_pairs['nodes_values'].append(value)
	attr_pairs['nodes_truenodeids'].append(true_child_id)
	attr_pairs['nodes_falsenodeids'].append(false_child_id)

def add_leaf(attr_pairs, tree_id, node_id, mode, weight):
	add_node(attr_pairs, tree_id, node_id, 0, mode, 0.0, 0, 0)

	attr_pairs['target_treeids'].append(tree_id)
	attr_pairs['target_nodeids'].append(node_id)
	attr_pairs['target_ids'].append(0)
	attr_pairs['target_weights'].append(weight)

def add_tree_to_attribute_pairs(attr_pairs, tree, tree_id, model):
	parents = {}

	for i in range(tree.node_count):
		node_id = i

		if tree.children_left[i] > i or tree.children_right[i] > i:
			mode = 'BRANCH_LEQ'
			feat_id = tree.feature[i]
			threshold = tree.threshold[i]
			left_child_id = int(tree.children_left[i])
			right_child_id = int(tree.children_right[i])
			parents[left_child_id] = i
			parents[right_child_id] = i

			add_node(attr_pairs, tree_id, node_id, feat_id, mode, threshold, left_child_id, right_child_id)
		else:
			mode = 'LEAF'
			weight = get_leaf_weight(tree_id, node_id, parents, model)

			add_leaf(attr_pairs, tree_id, node_id, mode, weight)


def convert_aad_forest(scope, operator, container):
	model = operator.raw_operator

	attr_pairs = get_default_attribute_pairs()

	for i, tree in enumerate(model.clf.estimators_):
		add_tree_to_attribute_pairs(attr_pairs, tree.tree_, i, model)

	container.add_node('TreeEnsembleRegressor', operator.input_full_names,
		operator.output_full_names, op_domain='ai.onnx.ml',
		name=scope.get_unique_operator_name('TreeEnsembleRegressor'), **attr_pairs)

register_converter('AadForest', convert_aad_forest)



from onnxconverter_common.registration import register_shape_calculator
from onnxconverter_common.data_types import FloatTensorType
from onnxconverter_common.utils import check_input_and_output_types, check_input_and_output_numbers

def calculate_aad_forest_output_shapes(operator):
	check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
	check_input_and_output_types(operator, good_input_types=[FloatTensorType])
	N = operator.inputs[0].type.shape[0]

	operator.outputs[0].type = FloatTensorType(shape=[N])

register_shape_calculator('AadForest', calculate_aad_forest_output_shapes)



from ._container import AadModelContainer
from onnxconverter_common.topology import Topology
from onnxconverter_common.data_types import FloatTensorType

from ad_examples.aad.forest_aad_detector import AadForest

def _get_aad_operator_name(model):
	# FIXME: not all possible AAD models are currently supported
	if not isinstance(model, AadForest):
		raise ValueError("No proper operator name found for '%s'" % type(model))

	return "AadForest"

def _parse_aad(scope, model, inputs):
	this_operator = scope.declare_local_operator(_get_aad_operator_name(model), model)
	this_operator.inputs = inputs

	# FIXME: probably another variable is required for anomality label
	score_variable = scope.declare_local_variable('score', FloatTensorType())

	this_operator.outputs.append(score_variable)

	return this_operator.outputs

def parse_aad(model, initial_types=None, target_opset=None,
	custom_conversion_functions=None, custom_shape_calculators=None):

	raw_model_container = AadModelContainer(model)
	topology = Topology(raw_model_container, default_batch_size='None',
		initial_types=initial_types, target_opset=target_opset,
		custom_conversion_functions=custom_conversion_functions,
		custom_shape_calculators=custom_shape_calculators)
	scope = topology.declare_scope('__root__')

	inputs = []
	for var_name, initial_type in initial_types:
		inputs.append(scope.declare_local_variable(var_name, initial_type))

	for variable in inputs:
		raw_model_container.add_input(variable)

	outputs = _parse_aad(scope, model, inputs)

	for variable in outputs:
		raw_model_container.add_output(variable)

	return topology



from setuptools import setup, find_packages

name="aad2onnx"
version="0.1.1"
test_suite="tests"

setup(name=name,
	version=version,
	description='Python module to convert AAD model to ONNX format',
	url='http://github.com/matwey/aad2onnx',
	author='Matwey V. Kornilov',
	author_email='matwey.kornilov@gmail.com',
	packages=find_packages(exclude=(test_suite,)),
	test_suite=test_suite,
	license='MIT',
	install_requires=[
		'onnx',
		'onnxconverter-common >= 1.8.0',
		'onnxruntime',
		'skl2onnx',
		'ad_examples @ git+https://github.com/shubhomoydas/ad_examples.git@0a7b86c4f2f7306ff543a15b387fe938f9c06130#egg=ad_examples-0.0.1'
	],
	classifiers=[
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
	],
	zip_safe=False)
